"""
background.py 

Sept 17, 2020
"""

import itertools
import math
import numpy as np
import os
import pandas as pd
import sys
import warnings

from collections import Counter
from tcrdist.pgen import OlgaModel
from tcrsampler.sampler import TCRsampler

from tcrdist.automate import auto_pgen


flatten = lambda l: [item for sublist in l for item in sublist]

def make_gene_usage_counter(df, cols = ['v_b_gene', 'j_b_gene']):
    """
    >>>
    """
    gene_usage_list= df[cols].to_dict('split')['data']
    gene_usage_list = [(x[0],x[1]) for x in gene_usage_list]
    gene_usage_counter = Counter(gene_usage_list )
    return(gene_usage_counter)



######################
#### CONSTRUCTION ####
######################

def sim_all_cdr3_gen(n = 100, recomb_type="VDJ", chain_folder = "human_T_beta", cols = ['v_b_gene', 'j_b_gene', 'cdr3_b_aa']):

    def expand_grid(dct):
        rows = itertools.product(*dct.values())
        return pd.DataFrame.from_records(rows, columns=dct.keys())
    
    omb = OlgaModel(recomb_type=recomb_type, chain_folder = chain_folder)
    
    all_vs = [x for x in omb.pgen_model.V_allele_names if x.endswith('*01')]
    all_js = [x for x in omb.pgen_model.J_allele_names if x.endswith('*01')]
    all_possible_beta = expand_grid(
        {'V': all_vs ,
         'J': all_js }
    )


    find_nones = list()
    results = list()
    
    for i,r in all_possible_beta.iterrows():
        
        e = omb.gen_cdr3s(V = r['V'], J = r['J'], n = n)
        results.append(pd.DataFrame({cols[2] : e,  cols[0]: r['V'], cols[1]:r['J']}))
    
        if e[0] is None:
            find_nones.append( [r['V'],r['J']])
 
    return results, find_nones

# Using 4.6M TCRs; Determine frequencies or subsitute minimum frequences

def get_gene_frequencies(ts, df, cols = ['v_b_gene','j_b_gene']):
    min_pV = np.min(list(ts.v_occur_freq.values()))
    min_pJ = np.min(list(ts.j_occur_freq.values()))
    min_pVJ = np.min(list(ts.vj_occur_freq.values()))
    df['pV'] = df[cols[0]].apply(lambda x : ts.v_occur_freq.get(x, min_pV))
    df['pJ'] = df[cols[1]].apply(lambda x : ts.j_occur_freq.get(x, min_pJ))
    df['pVJ'] = [ts.vj_occur_freq.get((r[cols[0]], r[cols[1]]), min_pVJ) for i,r in df[cols].iterrows()] 
    return df


def calculate_adjustment(df, cols = ['v_b_gene','j_b_gene'], adjcol = 'pVJ'):
    """
    
    Returns
    -------
    pd.Series
    """
    xdf = df[cols].groupby(cols).size().reset_index()
    xdf.columns = cols + ['n']
    # For each V,J pairing compute frequency in this reference
    xdf['ref_freq'] = xdf['n'] / xdf['n'].sum()
    df = df.merge(xdf, how = "left", on = cols).reset_index()
    return  pd.Series(df[adjcol] / df['ref_freq'])




def make_flat_vj_background(ts = None, n=200 , size =100000  ,  cols = ['v_b_gene','j_b_gene']):
    """
    Parameters
    ----------
    ts : TCRsampler
    
    n : int
        Default 200, number of TCRs to generate using Olga
    cols : list
        Default ['v_b_gene','j_b_gene']

    Returns 
    -------
    df : DataFrame

    Makes a flat background where every V,J pair is equally represeented. 

    Enrichment factors for each pair are based on frequency distribution
    of VJ pairing in a TCRsamler
    """
    if size/n < 135:
        raise ValueError(f"Based on size = {size}, increase to alteast {size/1000} to have sufficient TCRs per VJ pairing") 
    if ts is None:
        from tcrsampler.sampler import TCRsampler
        ts = TCRsampler(default_background = 'britanova_human_beta_t_cb.tsv.sampler.tsv')

    results, find_nones = sim_all_cdr3_gen(n = n)
    dfopt = pd.concat(results)
    dfopt = dfopt[dfopt.cdr3_b_aa.notna()]
    # import numpy as np
    # min_pV = np.min(list(ts.v_occur_freq.values()))
    # min_pJ = np.min(list(ts.j_occur_freq.values()))
    # min_pVJ = np.min(list(ts.vj_occur_freq.values()))
    # dfopt['pV'] = dfopt.v_b_gene.apply(lambda x : ts.v_occur_freq.get(x, min_pV))
    # dfopt['pJ'] = dfopt.j_b_gene.apply(lambda x : ts.j_occur_freq.get(x, min_pJ))
    # dfopt['pVJ'] = [ts.vj_occur_freq.get((r[cols[0]], r[cols[1]]), min_pVJ) for i,r in dfopt[cols].iterrows()] 

    min_n = dfopt.groupby(cols).size().min()
    import math 
    n = math.ceil(size / dfopt.groupby(cols).size().shape[0])
    min_n = min(min_n, n)
    parts = list()
    for i,g in dfopt.groupby(cols):
        parts.append(g.sample(min_n))
    df = pd.concat(parts).reset_index(drop = True)
    #df.to_csv("olga_optimized_human_T_beta.csv", index = False)
    df = get_gene_frequencies(ts = ts, df = df, cols = cols)
    return df



def make_vj_matched_background(
    gene_usage_counter, 
    ts = None, 
    size = 100000, 
    recomb_type="VDJ", 
    chain_folder = "human_T_beta",
    cols = ['v_b_gene', 'j_b_gene', 'cdr3_b_aa']):
    """
    gene_usage_counter : collections.Counter
    size : int
    recomb_type : str
        Default "VDJ", 
    chain_folder : str
        Default is for human beta "human_T_beta",
    cols : list 
        Default is for beta ['v_b_gene', 'j_b_gene', 'cdr3_b_aa']
    
    Example
    -------
    >>> ix =[['TRBV19*01', 'TRBJ2-5*01', 3],['TRBV24-1*01', 'TRBJ2-4*01', 3]]
    >>> df_rare= pd.concat([pd.DataFrame({'cdr3_b_aa' : flatten(ts.sample([[x[0], x[1], x[2]]])) , 'v_b_gene':x[0], 'j_b_gene':x[1]}) for x in ix]).reset_index(drop = True)
    >>> gene_usage_counter = make_gene_usage_counter(df_rare)
    >>> make_vj_matched_background(gene_usage_counter, size = 10)
          v_b_gene    j_b_gene            cdr3_b_aa        pV        pJ       pVJ
    0  TRBV24-1*01  TRBJ2-4*01      CATPVAGVAKNIQYF  0.011942  0.042163  0.000420
    1  TRBV24-1*01  TRBJ2-4*01       CATSPRGSLSIQYF  0.011942  0.042163  0.000420
    2  TRBV24-1*01  TRBJ2-4*01    CATSDLGGGGIHNIQYF  0.011942  0.042163  0.000420
    3    TRBV19*01  TRBJ2-5*01    CASSISDRGKFSETQYF  0.006788  0.089505  0.000394
    4  TRBV24-1*01  TRBJ2-4*01    CATSDLPARTRENIQYF  0.011942  0.042163  0.000420
    5  TRBV24-1*01  TRBJ2-4*01      CATSDPQGAKNIQYF  0.011942  0.042163  0.000420
    6    TRBV19*01  TRBJ2-5*01  CASSISCGRNLGGQETQYF  0.006788  0.089505  0.000394
    7    TRBV19*01  TRBJ2-5*01    CASSCKPSGGYQETQYF  0.006788  0.089505  0.000394
    8    TRBV19*01  TRBJ2-5*01     CASSSGTSHKLETQYF  0.006788  0.089505  0.000394
    9    TRBV19*01  TRBJ2-5*01          CASSDRETQYF  0.006788  0.089505  0.000394
    """
    
    olga_model_beta = OlgaModel(recomb_type=recomb_type, chain_folder = chain_folder)
    total_seqs = np.sum(list(gene_usage_counter.values()))
    adjust_factor = size / total_seqs
    
    dfs = list()
    adjust_depth = 1
    for k,v in gene_usage_counter.items():
        try:
            cdr3s = olga_model_beta.gen_cdr3s(V = k[0], J = k[1], n = v * math.ceil(adjust_factor))
            df = pd.DataFrame({cols[2]: cdr3s})
            df[cols[0]] = k[0]
            df[cols[1]] = k[1]
            dfs.append(df)
        except AttributeError:
            pass
    
    df = pd.concat(dfs).reset_index(drop = True)
    df = df[df[cols[2]].notna()][cols]
    
    if ts is None:
        from tcrsampler.sampler import TCRsampler
        ts = TCRsampler(default_background = 'britanova_human_beta_t_cb.tsv.sampler.tsv')
        ts = get_stratified_gene_usage_frequency(ts, replace = True)
    df = get_gene_frequencies(ts = ts, df = df, cols = cols)
    df = df.reset_index(drop = True)
    return(df)


def get_stratified_gene_usage_frequency(ts = None, replace = True):
    """
    MODIFIES A TCRsampler instance with esitmates vj_occur_freq_stratified by subject
    
    Parameters
    ----------
    ts : tcrsampler.sampler.TCRsampler

    replace : bool 
        if True, ts.v_occur_freq is set to ts.v_occur_freq_stratified 
        so other functions will work as befor.

    Returns
    -------
    ts : tcrsampler.sampler.TCRsampler

    """

    if ts is None:
        ts = TCRsampler(default_background = 'britanova_human_beta_t_cb.tsv.sampler.tsv')

    # (1/uniqueTCR_sample_depth) / nsubject
    nsubjects = len(ts.ref_df.subject.value_counts())
    inverse_tcrs_per_subject = (1/ts.ref_df.subject.value_counts()) / nsubjects
    # <weights df> 
    ws_df = pd.DataFrame({'subject': inverse_tcrs_per_subject.index, 'sweight': inverse_tcrs_per_subject}).reset_index(drop = True)
    # left join <ws_df> to provide a subject specific weight 
    df = ts.ref_df.merge(ws_df, how = 'left', on = 'subject').copy()
    # All sweights should sum to 1.0, up to rounding error
    assert np.isclose(df.sweight.sum() ,1.0)

    # SUBJECT STRATIFIED V,J FREQUENCIES
    # For each V,J combo take the weighted sum across all samples
    df_vj_occur_freq = df[['sweight','v_reps','j_reps']].groupby(['v_reps','j_reps']).sum().reset_index().rename(columns = {'sweight': 'pVJ'})
    assert np.isclose(df_vj_occur_freq.pVJ.sum() ,1.0)
    df_vj_occur_freq
    # Covert to a dictionary keyed on (V,J)
    ts.vj_occur_freq_stratified = { (x[0],x[1]): x[2] for x in df_vj_occur_freq.to_dict('split')['data']}

    # SUBJECT STRATIFIED VFREQUENCIES
    df_v_occur_freq = df[['sweight','v_reps']].groupby(['v_reps']).sum().reset_index().rename(columns = {'sweight': 'pV'})
    assert np.isclose(df_v_occur_freq.pV.sum() ,1.0)
    df_v_occur_freq
    # Covert to a dictionary keyed on (V,J)
    ts.v_occur_freq_stratified = { x[0]:x[1] for x in df_v_occur_freq.to_dict('split')['data']}

    # SUBJECT STRATIFIED JFREQUENCIES
    df_j_occur_freq = df[['sweight','j_reps']].groupby(['j_reps']).sum().reset_index().rename(columns = {'sweight': 'pJ'})
    assert np.isclose(df_j_occur_freq.pJ.sum() ,1.0)
    df_j_occur_freq
    # Covert to a dictionary keyed on (V,J)
    ts.j_occur_freq_stratified = { x[0]:x[1] for x in df_j_occur_freq.to_dict('split')['data']}
    
    if replace:
        warnings.warn("REPLACING ts.vj_occur_freq WITH ts.vj_occur_freq_stratified", stacklevel=2)
        warnings.warn("REPLACING ts.v_occur_freq  WITH ts.v_occur_freq_stratified", stacklevel=2)
        warnings.warn("REPLACING ts.j_occur_freq  WITH ts.j_occur_freq_stratified", stacklevel=2)
        ts.vj_occur_freq = ts.vj_occur_freq_stratified 
        ts.v_occur_freq  = ts.v_occur_freq_stratified 
        ts.j_occur_freq  = ts.j_occur_freq_stratified 

    return ts 


def sample_britanova(size = 100000, random_state =24082020):
    """
    Produce a background, stratfied by 8 subjects up to 960,000 TCR clones. 

    Unique TCRs are returned without consideration of their clonal frequency.

    Parameters
    ----------
    size : int 
        Size of background
    random_state : int
        Seed for random. sample
    """
    df = _get_britanova_human_beta_chord_blood_subject_stratified_background(size = size , random_state =random_state)
    return df

def _get_britanova_human_beta_chord_blood_subject_stratified_background(size = 100000, random_state =24082020):
    """
    Produce a background, stratfied by 8 subjects up to 960,000 TCR clones. 

    Unique TCRs are returned without consideration of their clonal frequency.

    Parameters
    ----------
    size : int 
        Size of background
    random_state : int
        Seed for random. sample
    """
    
    """Check for background file. If not present, download"""
    if not 'britanova_human_beta_t_cb.tsv.sampler.tsv' in TCRsampler.currently_available_backgrounds():
        TCRsampler.download_background_file('britanova_human_beta_t_cb.tsv.sampler.tsv.zip')
    else:
        pass 
        # print("CONGRATS 'britanova_human_beta_t_cb.tsv.sampler.tsv' ALREADY INSTALLED")

    ts = TCRsampler(default_background='britanova_human_beta_t_cb.tsv.sampler.tsv')
    ts = get_stratified_gene_usage_frequency(ts = ts, replace = True)
    # In [10]: ts.ref_df.subject.value_counts()
    # Out[10]:
    # A5-S18.txt    1073416
    # A5-S17.txt     825507
    # A5-S13.txt     692050
    # A5-S12.txt     573373
    # A5-S16.txt     559980
    # A5-S11.txt     519582
    # A5-S14.txt     302288
    # A5-S15.txt     120302 (NOTE THIS IS THE SMALLED STAMPLE)

    total = size  #100K
    nsubject = 8
    import math
    per_sample = math.ceil( total / nsubject)
    if per_sample > 120000:
        raise ValueError("Size: {size} exceed max size (960000) for valid stratification based on smallest sample")

    samples = []
    for subject_name, subject_df in ts.ref_df.groupby('subject'):
        if subject_name == 'A5-S15.txt':
            samples.append(subject_df.sample(per_sample, 
                replace = False,
                random_state = random_state).copy().reset_index(drop = True))
        else:
            samples.append(subject_df.sample(per_sample, 
                replace = False, 
                random_state = random_state).copy().reset_index(drop = True))

    bitanova_unique_clones_sampled = pd.concat(samples).reset_index(drop = True)
    bitanova_unique_clones_sampled = bitanova_unique_clones_sampled[['v_reps', 'j_reps', 'cdr3']].rename(columns = {'v_reps':'v_b_gene', 'j_reps':'j_b_gene','cdr3':'cdr3_b_aa'})
    return bitanova_unique_clones_sampled 


def _synthesize_human_beta_vj_background(ts,fn = None, df = None, n = 100000):
    """
    _build_vj_background

    Parameters
    ----------
    ts: tcrsampler.TCRsampler() 
        a TCRsampler instance, with gene usage frequencies (ideally computed get_stratified_gene_usage_frequency()
    fn: str
        file path to MIRA set of TCRs
    df : pandas DataFrame
        MIRA set of TCRs
    n : int 
        size of background to return
    Returns
    -------
    df_vj_bkgd : Pandas DataFrame
        A set of background TCRs with the same V and J gene usage as the input set. 
        These are generated using OLGA (Sethna et al.)
    """
    if fn is not None and df is not None:
        raise ValueError("_build_vj_background can accept <df> or <fn> arguments but not both")
    if fn is not None:
        # Load a set set of TCRs.
        df_target = pd.read_csv(fn)
    if df is not None:
        df_target = df.copy()
    # Subset columns
    df_target = df_target[['v_b_gene','j_b_gene','cdr3_b_aa']]
    # Make a gene usage counter
    gene_usage_counter = make_gene_usage_counter(df_target)                             # 2
    print("MAKING A V-GENE, J-GENE MATCHED BACKGROUND.")
    # Check that sampler is using sample stratified frequencies. 
    assert ts.v_occur_freq is ts.v_occur_freq_stratified
    print("USING STRATIFIED FREQUENCIES.")
    # Make a vj matched background. 
    #   Note: <size> aregument should be greater than desired, because Olga can return none due to non-productive CDR3s.
    df_vj_bkgd = make_vj_matched_background(ts = ts,
        gene_usage_counter = gene_usage_counter,    
        size = int(1.5 * n), # we make more just in case some are missing 
        recomb_type="VDJ", 
        chain_folder = "human_T_beta",
        cols = ['v_b_gene', 'j_b_gene', 'cdr3_b_aa'])
    # Sample to get the desired number of TCRs from teh v,j matched set
    df_vj_bkgd = df_vj_bkgd.sample(n, random_state = 1).reset_index(drop = True)
    print("CALCULATE INVERSE PROBABILITY WEIGHT ADJUSTMENT.")
    # Calculate the invese weighting adjustmetn
    df_vj_bkgd['weights'] = calculate_adjustment(df = df_vj_bkgd, adjcol = "pVJ")
    df_vj_bkgd['source'] = "vj_matched"
    # Combine
    return df_vj_bkgd

def _synthesize_human_alpha_vj_background(ts,fn = None, df = None, n=100000):
    """
    _build_vj_background

    Parameters
    ----------
    ts: tcrsampler.TCRsampler() 
        a TCRsampler instance, with gene usage frequencies (ideally computed get_stratified_gene_usage_frequency()
    fn: str
        file path to MIRA set of TCRs
    df : pandas DataFrame
        MIRA set of TCRs
    Returns
    -------
    df_vj_bkgd : Pandas DataFrame
        A set of background TCRs with the same V and J gene usage as the input set. 
        These are generated using OLGA (Sethna et al.)
    """
    if fn is not None and df is not None:
        raise ValueError("_build_vj_background can accept <df> or <fn> arguments but not both")
    if fn is not None:
        # Load a set set of TCRs.
        df_target = pd.read_csv(fn)
    if df is not None:
        df_target = df.copy()
    # Subset columns
    df_target = df_target[['v_a_gene','j_a_gene','cdr3_a_aa']]
    # Make a gene usage counter
    gene_usage_counter = make_gene_usage_counter(df_target, cols = ['v_a_gene', 'j_a_gene'])                             # 2
    print("MAKING A V-GENE, J-GENE MATCHED BACKGROUND.")
    # Check that sampler is using sample stratified frequencies. 
    assert ts.v_occur_freq is ts.v_occur_freq_stratified
    print("USING STRATIFIED FREQUENCIES.")
    # Make a vj matched background. 
    #   Note: <size> aregument should be greater than desired, because Olga can return none due to non-productive CDR3s.
    df_vj_bkgd = make_vj_matched_background(ts = ts,
        gene_usage_counter = gene_usage_counter,    
        size = int(n * 1.5),
        recomb_type="VJ", 
        chain_folder = "human_T_alpha",
        cols = ['v_a_gene', 'j_a_gene', 'cdr3_a_aa'])
    # Sample to get the desired number of TCRs from teh v,j matched set
    df_vj_bkgd = df_vj_bkgd.sample(n, random_state = 1).reset_index(drop = True)
    print("CALCULATE INVERSE PROBABILITY WEIGHT ADJUSTMENT.")
    # Calculate the invese weighting adjustmetn
    df_vj_bkgd['weights'] = calculate_adjustment(df = df_vj_bkgd, adjcol = "pVJ", cols = ['v_a_gene','j_a_gene'])
    df_vj_bkgd['source'] = "vj_matched"
    # Combine
    return df_vj_bkgd


def _synthesize_mouse_beta_vj_background(ts, fn = None, df = None, n=100000):
    """
    _build_vj_background

    Parameters
    ----------
    ts: tcrsampler.TCRsampler() 
        a TCRsampler instance, with gene usage frequencies (ideally computed get_stratified_gene_usage_frequency()
    fn: str
        file path to MIRA set of TCRs
    df : pandas DataFrame
        MIRA set of TCRs
    Returns
    -------
    df_vj_bkgd : Pandas DataFrame
        A set of background TCRs with the same V and J gene usage as the input set. 
        These are generated using OLGA (Sethna et al.)
    """
    if fn is not None and df is not None:
        raise ValueError("_build_vj_background can accept <df> or <fn> arguments but not both")
    if fn is not None:
        # Load a set set of TCRs.
        df_target = pd.read_csv(fn)
    if df is not None:
        df_target = df.copy()
    # Subset columns
    df_target = df_target[['v_b_gene','j_b_gene','cdr3_b_aa']]
    # Make a gene usage counter
    gene_usage_counter = make_gene_usage_counter(df_target, cols = ['v_b_gene', 'j_b_gene'])                             # 2
    print("MAKING A V-GENE, J-GENE MATCHED BACKGROUND.")
    # Check that sampler is using sample stratified frequencies. 
    assert ts.v_occur_freq is ts.v_occur_freq_stratified
    print("USING STRATIFIED FREQUENCIES.")
    # Make a vj matched background. 
    #   Note: <size> aregument should be greater than desired, because Olga can return none due to non-productive CDR3s.
    df_vj_bkgd = make_vj_matched_background(ts = ts,
        gene_usage_counter = gene_usage_counter,    
        size = int(n * 1.5),
        recomb_type="VDJ", 
        chain_folder = "mouse_T_beta",
        cols = ['v_b_gene', 'j_b_gene', 'cdr3_b_aa'])
    # Sample to get the desired number of TCRs from teh v,j matched set
    df_vj_bkgd = df_vj_bkgd.sample(n, random_state = 1).reset_index(drop = True)
    print("CALCULATE INVERSE PROBABILITY WEIGHT ADJUSTMENT.")
    # Calculate the invese weighting adjustmetn
    df_vj_bkgd['weights'] = calculate_adjustment(df = df_vj_bkgd, adjcol = "pVJ")
    df_vj_bkgd['source'] = "vj_matched"
    # Combine
    return df_vj_bkgd

def _synthesize_mouse_alpha_vj_background(ts, fn = None, df = None, n=100000):
    """
    _build_vj_background

    Parameters
    ----------
    ts: tcrsampler.TCRsampler() 
        a TCRsampler instance, with gene usage frequencies (ideally computed get_stratified_gene_usage_frequency()
    fn: str
        file path to MIRA set of TCRs
    df : pandas DataFrame
        MIRA set of TCRs
    Returns
    -------
    df_vj_bkgd : Pandas DataFrame
        A set of background TCRs with the same V and J gene usage as the input set. 
        These are generated using OLGA (Sethna et al.)
    """
    if fn is not None and df is not None:
        raise ValueError("_build_vj_background can accept <df> or <fn> arguments but not both")
    if fn is not None:
        # Load a set set of TCRs.
        df_target = pd.read_csv(fn)
    if df is not None:
        df_target = df.copy()
    # Subset columns
    df_target = df_target[['v_a_gene','j_a_gene','cdr3_a_aa']]
    # Make a gene usage counter
    gene_usage_counter = make_gene_usage_counter(df_target, cols = ['v_a_gene', 'j_a_gene'] )                             # 2
    print("MAKING A V-GENE, J-GENE MATCHED BACKGROUND.")
    # Check that sampler is using sample stratified frequencies. 
    assert ts.v_occur_freq is ts.v_occur_freq_stratified
    print("USING STRATIFIED FREQUENCIES.")
    # Make a vj matched background. 
    #   Note: <size> aregument should be greater than desired, because Olga can return none due to non-productive CDR3s.
    df_vj_bkgd = make_vj_matched_background(ts = ts,
        gene_usage_counter = gene_usage_counter,    
        size = int(n * 1.5),
        recomb_type="VJ", 
        chain_folder = "mouse_T_alpha",
        cols = ['v_a_gene', 'j_a_gene', 'cdr3_a_aa'])
    # Sample to get the desired number of TCRs from teh v,j matched set
    df_vj_bkgd = df_vj_bkgd.sample(n, random_state = 1, replace = True).reset_index(drop = True)
    print("CALCULATE INVERSE PROBABILITY WEIGHT ADJUSTMENT.")
    print(df_vj_bkgd)
    # Calculate the invese weighting adjustmetn
    df_vj_bkgd['weights'] = calculate_adjustment(df = df_vj_bkgd, adjcol = "pVJ", cols = ['v_a_gene','j_a_gene'])
    df_vj_bkgd['source'] = "vj_matched"
    # Combine
    return df_vj_bkgd



import pytest

def test_quick_pipeline_with_fragmented_compute():

	"""
	How can I used tcrdist3 to test for TCRs that may HLA restricted. 

	
	"""

	import os
	import pandas as pd
	import numpy as np
	from scipy import sparse
	from tcrdist.repertoire import TCRrep
	from tcrdist.rep_funcs import  compute_pw_sparse_out_of_memory
	
	f = 'mira_epitope_67_382_APHGVVFL_APHGVVFLHV_GVVFLHVTY_VVFLHVTYV.tcrdist3.csv'
	f = os.path.join('tcrdist','data','covid19',f)
	assert os.path.isfile(f)

	df = pd.read_csv(f)
	df = df[['subject', 'cell_type', 'v_b_gene', 'j_b_gene', 'cdr3_b_aa', 'cdr3_b_nucseq',  'cohort', 'hla-a', 'hla-a_1','hla-b', 'hla-b_1']]
	tr = TCRrep(cell_df = df,               
				organism = 'human',
				chains = ['beta'],
				db_file = 'alphabeta_gammadelta_db.tsv',
				compute_distances = False,
				store_all_cdr = False)

	from tcrdist.rep_funcs import  compute_pw_sparse_out_of_memory
	
	S, fragments = compute_pw_sparse_out_of_memory(	tr = tr,
													row_size      = 100,
													pm_processes  = 2,
													pm_pbar       = True,
													max_distance  = 1000,
													matrix_name   = 'rw_beta',
													reassemble    = True,
													cleanup       = False)

	tr.clone_df['B07'] = (tr.clone_df['hla-b'].str.startswith("B*07") | tr.clone_df['hla-b_1'].str.startswith("B*07"))
	tr.clone_df['B07'] = ["B*07" if (x) else "NOTB*07 " for x in tr.clone_df['B07']]

	#sparse.save_npz("S.npz", S)
	from tcrdist.rep_funcs import  compute_n_tally_out_of_memory
	nn_tally_df_cohort = compute_n_tally_out_of_memory(fragments,
												matrix_name = "rw_beta",
												pm_processes  = 6,
												to_file = False,
												to_memory = True, 
												knn_radius = 25, 
												x_cols = ['B07'])

	from hierdiff.association_testing import cluster_association_test
	nn_associations = cluster_association_test(res = nn_tally_df_cohort, y_col='cmember', method='fishers')
	nn_associations = nn_associations.sort_values('pvalue', ascending = True)
	import ast 
	nn_associations['neighbors_i'] = nn_associations.neighbors.apply(lambda x: ast.literal_eval(x))

	from tcrdist.summarize import test_for_almost_subsets, filter_is, filter_gt
	nn_associations['mostly_unique'] = test_for_almost_subsets(nn_associations['neighbors_i'], thr = 5)
	nr_nn_associations = filter_is(nn_associations, 'mostly_unique', 1).copy()

	#nr_nn_associations = filter_gt(nr_nn_associations, 'K_neighbors', 25).copy()
	nr_nn_associations


	# MOTIF GENERATION
	from tcrsampler.sampler import TCRsampler
	t = TCRsampler()
	if  'olga_human_beta_t.sampler.tsv' not in t.currently_available_backgrounds():
		t.download_background_file('olga_sampler.zip')
	#t.download_background_file('olga_sampler.zip') # ONLY IF NOT ALREADY DONE
	tcrsampler_beta = TCRsampler(default_background = 'olga_human_beta_t.sampler.tsv')
	tcrsampler_beta.build_background(max_rows = 1000)

	"""SEE PALMOTIF DOCS (https://github.com/agartland/palmotif)"""
	from palmotif import compute_pal_motif, svg_logo
	from tcrdist.summarize import _select
	
	"""GENERATE SVG GRAPHIC FOR EACH NODE OF THE TREE"""
	#pwmat_str = 'pw_beta'
	cdr3_name = 'cdr3_b_aa'
	gene_names = ['v_b_gene','j_b_gene']
	svgs_beta = list()
	svgs_beta_raw = list()
	info_list = list()

	from tcrdist.rep_diff import member_summ
	summary = member_summ(  res_df = nr_nn_associations,
							clone_df = tr.clone_df,
							addl_cols=['cohort','hla-a', 'hla-a_1', 'hla-b', 'hla-b_1', 'subject'])

	nr_nn_associations = pd.concat([nr_nn_associations, summary], axis = 1).reset_index()

	for i,r in nr_nn_associations.head(25).iterrows():
		dfnode  = tr.clone_df.iloc[r['neighbors_i'],:].copy()
		# <pwnode> Pairwise Matrix for node sequences
		pwnode = S[r['neighbors_i'],:] [:,r['neighbors_i']].todense()
		if dfnode.shape[0] > 2:
			iloc_idx = pwnode.sum(axis = 0).argmin()
			centroid = dfnode[cdr3_name].to_list()[iloc_idx]
		else:
			centroid = dfnode[cdr3_name].to_list()[0]

		print(f"CENTROID: {centroid}")

		gene_usage_beta = dfnode.groupby(gene_names).size()
		sampled_rep = tcrsampler_beta.sample( gene_usage_beta.reset_index().to_dict('split')['data'],
			flatten = True, depth = max(100, 1000 // dfnode.shape[0]))

		sampled_rep  = [x for x in sampled_rep if x is not None]

		motif, stat = compute_pal_motif(
						seqs = _select(df = tr.clone_df,
									   iloc_rows = r['neighbors_i'],
									   col = cdr3_name),
						refs = sampled_rep,
						centroid = centroid)

		svgs_beta.append(svg_logo(motif, return_str= True))

		sampled_rep = sampled_rep.append(centroid)
		motif_raw, _ = compute_pal_motif(
					 seqs =_select(df = tr.clone_df,
									iloc_rows = r['neighbors_i'],
									col = cdr3_name),
					 centroid = centroid)
		svgs_beta_raw.append(svg_logo(motif_raw, return_str= True))
		info_list.append(r)


	def row_to_string(r, vals = ['ct_columns', 'val_0', 'ct_0', 'val_1', 'ct_1', 'val_2', 'ct_2','val_3', 'ct_3', 'levels', 'K_neighbors', 'R_radius', 'RR', 'OR', 'pvalue', 'FWERp','FDRq']):
		#d = {v:r[v] for v in vals}
		return "<br></br>".join([f"\t{v} : {r[v]}" for v in vals])

	def to_html_table(r, vals = ['ct_columns', 'hla-a', 'hla-a_1', 'hla-b', 'hla-b_1', 'val_0', 'ct_0', 'val_2', 'ct_2', 'K_neighbors', 'R_radius', 'pvalue', 'FDRq','cdr3_b_aa','v_b_gene', 'j_b_gene', 'cohort','subject']):
		return pd.DataFrame(r[vals]).transpose().to_html()

	def shrink(html_str):
		return html_str.replace('height="100%"',  'height="10%"').\
			replace('width="100%"', 'width="10%"')

	with open('svgs_in_line.html', 'w') as fh:
		fh.write(f"<html><body>\n")
		

		for svg, svg_raw, details in zip(svgs_beta, svgs_beta_raw, info_list):
			fh.write(f"{shrink(svg_raw)}{shrink(svg)}")
			try:
				fh.write(to_html_table(details))
			except:
				print("F")
			fh.write("<div></div>")
		fh.write(f"</html></body>\n")






def test_ranked_centers_example():
    """
    Report of meta-clonotypes using two dataframes.
    <df>  has all TCRS
    <df2> has a subset of TCRS in <df>, specifiyint which 
    are to be used as centroids.
    """
    import os
    import pandas as pd
    import numpy as np
    from tcrdist.repertoire import TCRrep
    from tcrdist.public import TCRpublic  
    from tcrdist.tree import _default_sampler_olga
    from progress.bar import IncrementalBar
    from palmotif import compute_pal_motif, svg_logo
    from tcrdist.public import make_motif_logo

    output_html_name = "custom_report.html"
    # <fn> filename for all TCRs in an antigen-enriched repertoire
    fn = os.path.join('tcrdist','data','covid19',
        'mira_epitope_55_524_ALRKVPTDNYITTY_KVPTDNYITTY.tcrdist3.csv.bE5ctrl.centers.csv')
    df = pd.read_csv(fn, sep = ",")
    df = df[['cdr3_b_aa', 'v_b_gene', 'j_b_gene', 'pgen', 'max_radi']].\
        rename(columns= {'max_radi':'radius'}).copy()

    # <fn>2 filename for priority TCRs
    fn2 = os.path.join('tcrdist','data','covid19',
        'mira_epitope_55_524_ALRKVPTDNYITTY_KVPTDNYITTY.tcrdist3.csv.bE5ctrl.centers.csv.ranked_centers.tsv')
    df2 = pd.read_csv(fn2, sep = "\t").\
        rename(columns= {'max_radi':'radius'}).copy()

    # Compute distances between all TCRs
    tr = TCRrep(cell_df = df, 
        organism = 'human',
        chains = ['beta'], 
        compute_distances = True)

    # Initialize a tcrsampler, this will be used to make background motifs
    tcrsampler = _default_sampler_olga(organism = "human", chain = "beta")()

    # Iterate through each row of the df2, making a logo for each.
    svgs = list()
    svgs_raw = list()
    bar = IncrementalBar("Making Logos", max = df2.shape[0])
    for i,r in df2.iterrows():
        bar.next()
        svg,svg_raw=make_motif_logo(tcrsampler = tcrsampler,
                            clone_df = tr.clone_df,
                            pwmat = tr.pw_beta,
                            centroid = r['cdr3_b_aa'],
                            v_gene = r['v_b_gene'],
                            radius = r['radius'],
                            pwmat_str = 'pw_beta',
                            cdr3_name = 'cdr3_b_aa',
                            v_name = 'v_b_gene',
                            gene_names = ['v_b_gene','j_b_gene'])
        svgs.append(svg)
        svgs_raw .append(svg_raw)
    bar.next(); bar.finish()
    df2['svg'] = svgs
    df2['svg_raw'] = svgs_raw

    def shrink(s):
        """reduce size of svg graphic"""
        s = s.replace('height="100%"', 'height="20%"')
        s = s.replace('width="100%"', 'width="20%"')
        return s

    # Choose columns to include in the report
    labels = [  'cdr3_b_aa', 
                'v_b_gene',
                'j_b_gene',
                'radius',
                'regex', 
                'target_hits',
                'nsubject',
                'chi2joint']

    with open(output_html_name, 'w') as output_handle:
        for i,r in df2.iterrows():
            #import pdb; pdb.set_trace()
            svg, svg_raw = r['svg'],r['svg_raw']
            output_handle.write("<br></br>")
            output_handle.write(shrink(svg))
            output_handle.write(shrink(svg_raw))
            output_handle.write("<br></br>")
            output_handle.write(pd.DataFrame(r[labels]).transpose().to_html())
            output_handle.write("<br></br>")


